import re
import warnings

import AutoRPE.UtilsRPE.Error as Error
import AutoRPE.UtilsRPE.precision_module as precision_module
from AutoRPE.UtilsRPE.Functions import remove_comments, load_intrinsics, strip_call_arguments, close_brackets, \
    in_which_block, \
    is_keyword_argument, split_elements_in_operations, function_type_from_header, type_of_contents, find_call_arguments, \
    split_semicolon, literal_mask, find_kind_of_parameter_exception, add_rpe_to_argument, add_rpe_var_to_arguments, \
    find_called_routine, is_a_number, split_elements_in_comparison, replace_variable_exact_match, \
    calls_to_routines, calls_to_procedure, calls_to_functions, find_called_function, find_argument_intervals, \
    add_val_to_argument, add_real_to_argument, split_declaration_arguments, separate_line_container_and_arguments, \
    match_arguments_and_interface, has_operations, split_read_arguments, parse_attributes, use_temporal_variable, \
    remove_literals, remove_outer_brackets, has_intrinsics, list_of_intrinsics, is_call_to_function, \
    is_start_function_statement, \
    is_end_function_statement

single = "single"
double = "double"
working = precision_module.working_precision
real_types = [single, double]


# Source Database Classes
class Vault:
    def __init__(self):
        self.modules = []
        self.variables = []
        self.dictionary_of_variables = {}
        self.procedures = []
        self.procedures_dictionary = None
        self.routines = []
        self.functions = []
        self.interfaces = []
        self.derived_types = []
        self.intrinsics = load_intrinsics()
        self.indexed_variables = None

    def add_module(self, module):
        self.modules.append(module)

    def __getitem__(self, item):
        for module in self.modules:
            if module.name == item:
                return module

    def process_procedures(self):
        # Purpose: Add the procedures defined in the modules to the list of procedures and classify by type.
        for module in self.modules:
            for procedure in module.procedures:
                if procedure not in self.procedures:
                    self.procedures.append(procedure)
        for procedure in self.procedures:
            if isinstance(procedure, Routine):
                if procedure not in self.routines:
                    self.routines.append(procedure)
            elif isinstance(procedure, Function):
                if procedure not in self.functions:
                    self.functions.append(procedure)
            elif isinstance(procedure, Interface):
                if procedure not in self.interfaces:
                    self.interfaces.append(procedure)
            elif isinstance(procedure, DerivedType):
                if procedure not in self.derived_types:
                    self.derived_types.append(procedure)

        self.procedures_dictionary = {"all": {}, "functions": {}, "subroutines": {}, "interfaces": {}, "types": {}}
        for procedure in self.procedures:
            if not isinstance(procedure, ModuleMain):
                # Determine which kind of procedure are we dealing with
                if isinstance(procedure, Function):
                    procedure_group = "functions"
                elif isinstance(procedure, Routine):
                    procedure_group = "subroutines"
                elif isinstance(procedure, Interface):
                    procedure_group = "interfaces"
                elif isinstance(procedure, DerivedType):
                    procedure_group = "types"
                else:
                    raise AssertionError("Procedure type not found")

                # Add procedure to dictionary, in its own section and in the dictionary with all elements
                try:
                    self.procedures_dictionary[procedure_group][procedure.name.lower()]
                    self.procedures_dictionary["all"][procedure.name.lower()]
                    procedure_arr = self.procedures_dictionary[procedure_group][procedure.name.lower()]
                    if procedure not in procedure_arr:
                        procedure_arr.append(procedure)
                        self.procedures_dictionary[procedure_group][procedure.name.lower()] = procedure_arr
                        self.procedures_dictionary["all"][procedure.name.lower()] = procedure_arr
                    else:
                        print("Procedure %s defined more than once" % procedure.name)

                except KeyError:
                    self.procedures_dictionary[procedure_group][procedure.name.lower()] = [procedure]
                    self.procedures_dictionary["all"][procedure.name.lower()] = [procedure]

    def process_variables(self):
        # Purpose: Add the variables defined in the modules to the list of variables.
        if not self.procedures:
            self.process_procedures()
        for procedure in self.procedures:
            if isinstance(procedure, ModuleMain) \
                    or isinstance(procedure, Function) \
                    or isinstance(procedure, Routine) \
                    or isinstance(procedure, DerivedType):
                for variable in procedure.variables:
                    self.variables.append(variable)
                    if not variable.name.lower() in self.dictionary_of_variables:
                        self.dictionary_of_variables[variable.name.lower()] = [variable]
                    else:
                        self.dictionary_of_variables[variable.name.lower()] += [variable]

    def process_pointers(self):
        for var in self.variables:
            if var.pointer:
                if var.pointer.count("type"):
                    pointer_name = var.pointer.replace("type", "")
                    pointer_name = remove_outer_brackets(pointer_name)
                    search_array = self.derived_types
                elif var.pointer.count("class"):
                    # Same of type, but for compiler reason they have to be declared as classes
                    pointer_name = var.pointer.replace("class", "")
                    pointer_name = remove_outer_brackets(pointer_name)
                    search_array = self.derived_types
                elif var.pointer.count("procedure"):
                    pointer_name = var.pointer.replace("procedure", "")
                    pointer_name = remove_outer_brackets(pointer_name)
                    search_array = self.functions
                else:
                    # It is a pointer to real or a pointer to integer
                    continue
                pointed_obj = [obj for obj in search_array if obj.name == pointer_name]
                if len(pointed_obj) > 1:
                    raise Error.DuplicatedObject
                else:
                    var.pointer = pointed_obj
                continue

    def index_variables(self):
        if not self.variables:
            self.process_variables()
        self.indexed_variables = {}
        for variable in self.variables:
            self.indexed_variables[variable.id] = variable

    def get_variable(self, item):
        # Purpose: Retrieve a variable by name
        if not self.variables:
            self.process_variables()
        try:
            for variable in self.dictionary_of_variables[item.lower()]:
                yield variable
        except KeyError:
            pass

    def get_variable_by_id(self, var_id):
        # Purpose: Retrieve a variable by id

        # Few workarounds in case the vault does not have the indexed_variables
        try:
            self.indexed_variables
        except AttributeError:
            self.indexed_variables = None
        if not self.indexed_variables:
            self.index_variables()

        return self.indexed_variables[var_id]

    def get_function(self, item):
        # Purpose: Retrieve a function by name
        return self.get_procedure(item, procedure_type=Function)

    def get_procedure(self, item, procedure_type=None, module_name=None):
        # Purpose: Retrieve a procedure by name
        item = item.lower()
        # TODO: This routine is quite expensive (16%), could be boosted using dictionaries.
        if not self.procedures or not self.procedures_dictionary:
            self.process_procedures()

        new_approach = True

        if new_approach:
            if procedure_type is None:
                procedure_group = "all"
            elif procedure_type == Function:
                procedure_group = "functions"
            elif procedure_type == Routine:
                procedure_group = "subroutines"
            elif procedure_type == Interface:
                procedure_group = "interfaces"
            elif procedure_type == DerivedType:
                procedure_group = "types"
            else:
                raise AssertionError("Procedure type not found")

            try:
                procedure = self.procedures_dictionary[procedure_group][item]
                if len(procedure) > 1 and module_name is not None:
                    # this is safe until they start to define functions inside functions
                    return [proc for proc in procedure if proc.module.name == module_name][0]

                return self.procedures_dictionary[procedure_group][item][0]
            except KeyError:
                raise Error.ProcedureNotFound("Procedure with name %s not found" % item)

        else:
            for procedure in self.procedures:
                if procedure_type:
                    if procedure.name == item:
                        if isinstance(procedure, procedure_type):
                            return procedure
                elif procedure.name == item:
                    return procedure
            raise Error.ProcedureNotFound("Procedure with name %s not found" % item)

    def get_derived_type(self, item):
        # Purpose: Retrieve a derived type by name
        for dt in self.derived_types:
            if dt.name.lower() == item.lower():
                return dt

    def expand_use_statement(self):
        for module in self.modules:
            # use statement already stored
            used_modules = [m.uses for m in self.modules if m.name == module.name][0]

            # Store use statement of used modules
            additional_used_modules = []
            for um in used_modules:
                if um.uses:
                    additional_used_modules.extend(um.uses)

            # Remove duplicates
            additional_used_modules = set(additional_used_modules)

            # Remove input already in used statement
            additional_used_modules = [aum for aum in additional_used_modules if aum not in module.uses]

            # Look recursively at used modules
            while not sorted([aum.name for aum in additional_used_modules]) == sorted([mu.name for mu in module.uses]):
                used_modules = []
                for aum in additional_used_modules:
                    if aum.name not in [m.name for m in module.uses]:
                        module.uses.append(aum)
                    if aum.uses:
                        used_modules.extend(um.uses)

                additional_used_modules = set(used_modules + module.uses)

        # Now we have just to loop through the list to see which modules use mine
        for module in self.modules:
            for mu in module.uses:
                if module not in mu.used_by:
                    mu.used_by.append(module)

        # Add public var of used modules to availables var
        for module in self.modules:
            for used_module in module.uses:
                if len(used_module.variables):
                    for var in used_module.variables:
                        module.accessible_var.append(var)

    def expand_derived_types(self):
        for derived_type in self.derived_types:
            if derived_type.extended_structure is not None:
                # Save in the mother structure it extension
                struct_to_extend = [s for s in self.derived_types if s.name == derived_type.extended_structure][0]
                struct_to_extend.extension = derived_type
                # save in the extension the mother structure
                derived_type.extension = struct_to_extend

    def add_module_2var(self):
        # For all the variables in vault
        for var in self.variables:
            var_module = [m for m in self.modules if m.name == var.routine.module.name]
            var.module = var_module[0]
        # for all the module accessible var
        for module in self.modules:
            for var in module.accessible_var:
                var_module = [m for m in self.modules if m.name == var.routine.module.name]
                var.module = var_module[0]

    def generate_namelist_precisions(self, filename="namelist_precisions", exclude_double_precision=True):
        rpe_vars = [var for var in self.variables if var.type == "rpe_var"]
        assigned_rpe_vars = [var for var in rpe_vars if var.mutable]
        assigned_rpe_vars = sorted(assigned_rpe_vars, key=lambda x: x.id)
        line_text = "emulator_variable_precisions(%d) = %d\t! Variable:%24s\tRoutine:%22s\tModule:%16s\n"
        with open(filename, "w") as namelist:
            namelist.write("! namelist variable precisions\n&precisions\n")
            for var in assigned_rpe_vars:
                if var.precision == 52 and exclude_double_precision:
                    continue
                if var.precision is None:
                    continue
                try:
                    if var.untouchable:
                        continue
                except AttributeError:
                    pass
                namelist.write(line_text % (var.id, var.precision, var.name, var.routine.name, var.routine.module.name))
            namelist.write("/\n")

    def assign_ids_to_immutable_variables(self):
        # Loop over rpe-variables and assign an id to those that don't have any, setting them to mutable=False
        last_id = max([v.id for v in self.variables if v.mutable is True])

        counter = last_id + 1
        for variable in self.variables:
            if variable.type == "rpe_var":
                if variable.id is None:
                    variable.id = counter
                    variable.mutable = False
                    counter += 1


class Module:
    def __init__(self, _name):
        self.name = _name
        self.repo = None
        self.procedures = []
        self.variables = []
        self.precision = None
        self.uses = []
        self.used_by = []
        self.accessible_var = []

    def add_procedure(self, _routine):
        assert isinstance(_routine, Procedure)
        _routine.module = self
        self.procedures.append(_routine)

    def add_variable(self, _variable):
        assert isinstance(_variable, Variable)
        _variable.routine = self
        self.variables.append(_variable)

    def __getitem__(self, item):
        for _routine in self.procedures:
            if _routine.name == item:
                return _routine
        new_routine = Routine(None, item)
        new_routine.module = self
        self.repo.add_procedure(new_routine)
        self.procedures.append(new_routine)
        return self[item]

    def set_precision(self, precision):
        assert isinstance(precision, int), "Precision argument in set_precision must be an integer"
        for routine in self.procedures:
            routine.set_precision(precision)
        self.precision = precision

    def main(self):
        for procedure in self.procedures:
            if isinstance(procedure, ModuleMain):
                return procedure
        return None


class Current_block:
    def __init__(self, _Module_main=None):
        if _Module_main is None:
            self.routine = None
            self.module_name = ""
        else:
            self.routine = _Module_main
            self.module_name = _Module_main.module.name
        self.parent_block = None


class Procedure(object):
    def __init__(self, _name, module):
        self.name = _name
        self.module = module
        self.variables = []
        self.precision = None
        self.interface = None
        self.dummy_arguments = None
        self.uses = []
        self._accessible_variables = None

    def add_variable(self, _variable):
        try:
            assert isinstance(_variable, Variable)
        except AssertionError:
            pass
        _variable.routine = self
        self.variables.append(_variable)

    def get_variable_by_position(self, position):
        for variable in self.variables:
            if variable.is_dummy_argument:
                if variable.position == position:
                    return variable

    def get_variable_by_name(self, name):
        for variable in self.variables:
            if variable.name.lower() == name.lower():
                return variable

    def accessible_variables(self):
        if self._accessible_variables is None:
            variables = self.variables
            if self.module:
                variables += self.module.main().variables
                if self.module.uses:
                    for module in self.module.uses:
                        variables += module.main().variables
            self._accessible_variables = variables

        return self._accessible_variables

    def get_dummy_arguments(self):
        if self.dummy_arguments is None:
            number_of_dummy_arguments = len([True for var in self.variables if var.is_dummy_argument])
            self.dummy_arguments = [self.get_variable_by_position(i) for i in range(number_of_dummy_arguments)]
        return self.dummy_arguments


class Routine(Procedure):
    def __init__(self, _name, module):
        super(Routine, self).__init__(_name, module)


class Function(Procedure):
    def __init__(self, _name, module):
        super(Function, self).__init__(_name, module)
        self.type = None


class DerivedType(Procedure):
    def __init__(self, _name, module, _extended_structure=None, ):
        super(DerivedType, self).__init__(_name, module)
        self.extended_structure = _extended_structure
        self.extension = None

    def get_variable(self, item):
        for variable in self.variables:
            if variable.name.lower() == item.lower():
                return variable


class ModuleMain(Procedure):
    def __init__(self, module):
        super(ModuleMain, self).__init__("main", module)
        self.module = module


class Interface:
    def __init__(self, _name):
        self.name = _name
        self.module = None
        self.procedures = []

    def add_procedure(self, _procedure):
        assert isinstance(_procedure, Routine) or isinstance(_procedure, Function)

        self.procedures.append(_procedure)
        # Don't know it will be useful yet
        if not _procedure.interface:
            _procedure.interface = self
        elif isinstance(_procedure.interface, list):
            _procedure.interface.append(self)
        else:
            _procedure.interface = [_procedure.interface, self]


class Variable:
    def __init__(self, _routine, _name, _type, _dimension, _module=None):
        self.id = None
        self.module = _module
        self.routine = _routine
        self.name = _name
        self.type = _type
        if self.type == "real":
            print("A real has been declared: %s" % _name)
        self.dimension = _dimension
        self.is_dummy_argument = False
        self.intent = None
        self.is_optional = None
        self.pointer = None
        self.position = None
        self.allocatable = False
        self.precision = None
        self.is_parameter = None
        self.is_member_of = None
        self.mutable = None
        self.same_as = []

    def set_precision(self, precision):
        assert isinstance(precision, int), "Precision argument in set_precision must be an integer"
        self.precision = precision


# File manipulation Classes
class ReadSourceFile(object):
    def __init__(self, _filename):
        # Init variables
        self.file_path = _filename
        self.filename = self.file_path.split("/")[-1]
        self.text = None
        self.lines = None

        # Read file
        self.read_file(_filename)

    def read_file(self, _filename):
        try:
            with open(_filename) as f:
                self.text = f.read()
                self.lines = self.text.split("\n")
        except UnicodeDecodeError:
            with open(_filename, encoding="latin1") as f:
                self.text = f.read()
                self.lines = self.text.split("\n")

    def rebuild_text(self):
        self.text = "".join([l + "\n" for l in self.lines[:]])

    def save_modified_file(self, filepath):
        with open(filepath, "w") as output_file:
            output_file.write(self.text)


# Classes used in ParseRoutines.py
class ObtainSourcefileInfo(ReadSourceFile):
    # Definition: This class is used to load a sourcefile and parse the subroutines contained there
    # Input: string with path to sourcefile
    def __init__(self, _filename):
        super(ObtainSourcefileInfo, self).__init__(_filename)
        # Definition: Initializing the object reading the sourcefile
        # Input: string with path to sourcefile
        # I think that it is safe to collect the information using all the code in lower case.
        self.text = self.text.lower()
        self.objects = []
        self.module = None

    # Definition: Given a line number, looks for next line with "end subroutine"
    # Input: integer with line number
    # Output: integer with the line number of the next "end subroutine"
    def find_next_routine_end(self, _index):
        return self.find_next_end(_index, "end subroutine")

    # Definition: Given a line number, looks for next line with "end function"
    # Input: integer with line number
    # Output: integer with the line number of the next "end function"
    def find_next_function_end(self, _index):
        return self.find_next_end(_index, "end function")

    # Definition: Given a line number, looks for next line with "end interface"
    # Input: integer with line number
    # Output: integer with the line number of the next "end interface"
    def find_next_interface_end(self, _index):
        return self.find_next_end(_index, "end interface")

    # Definition: Given a line number, looks for next line with the string contained in end_statement argument.
    # Input:
    #       _index: integer with line number
    #       end_statement: string that indicates next end_statement
    # Output: integer with the line number of the next line containing the end_statement
    def find_next_end(self, _index, end_statement):
        # for _line_index, _line in enumerate(self.lines[_index:]):
        #     if _line.lower().count(end_statement):
        #         return _line_index
        # return len(self.lines)
        _i = _index + 1
        begin_statement = end_statement.replace('end', '')
        nested_declaration = 0
        while _i < len(self.lines):
            _l = self.lines[_i]
            if re.match('^\s*(' + begin_statement + ')', _l, re.I):
                nested_declaration += 1
            if _l.lower().count(end_statement):
                if nested_declaration == 0:
                    return _i
                else:
                    nested_declaration -= 1
            _i += 1

        return len(self.lines)

    # Definition: Given a line number, returns the subroutine name and the dummy arguments (in case they exist)
    # Input: integer with line number
    # Output: tuple with routine name and list of dummy arguments
    def get_dummy_arguments(self, _index):
        _line = self.lines[_index]
        # Workaround to solve the problem that appears when there's a function declaration with (kind=...)
        if _line.lower().count("function "):
            if _line.find("(") < _line.lower().find("function "):
                _line = _line[_line.lower().find("function "):]

        # If the subroutine declaration does not have any parenthesis, this routine does not have any argument
        if not _line.count("("):
            _name = _line.split()[1].strip()
            _dummy_arguments = []
        # If otherwise it has dummy arguments, find them
        elif _line.lower().count("result") or _line.lower().count("bind("):
            _name = _line[:_line.find("(")].split()[-1]
            _dummy_arguments = []
            _dummy_arguments += strip_call_arguments(close_brackets(_line))
            # I don't know if we should record the result variable with the other variables
            # result_start = _line.lower().find("result")
            # _dummy_arguments += strip_call_arguments(_line[result_start:])
        else:
            _name = _line[:_line.find("(")].split()[-1]
            _dummy_arguments = strip_call_arguments(_line)
        _dummy_arguments = [_x for _x in _dummy_arguments if _x.strip()]
        return _name, _dummy_arguments

    # Definition: Given the line numbers where a routine starts and ends,
    # look for variable declarations.
    # Input:
    #       -_routine_start: integer with line number where the routine starts
    #       -_routine_end: integer with line number where the routine ends
    # Output: list of strings containing declarations
    def find_routine_declarations(self, _routine_start, _routine_end):

        # Check if there is some statement inside
        # TODO find a better way of parsing this: actually we can take advantage of the fact that
        # Cases in which a routine contains other functions declarations.
        stop_index = [index
                      for index, line in enumerate(self.lines[_routine_start:_routine_end + 1])
                      if line.count("contains") or line.count("interface")]
        # A declaration found stop the search there
        if stop_index:
            _routine_end = _routine_start - 1 + stop_index[0]

        # Lines inside the routine
        _routine_lines = self.lines[_routine_start:_routine_end + 1]

        # Lines that are variable declarations
        _all_declaration = [_l.strip() for _l in _routine_lines if _l.count("::")]
        # _declared_var = [_dec.split('::')[1].strip() for _dec in _all_declaration]
        # _used_declarations = []
        # for index, variable in enumerate(_declared_var):
        #     if variable.count("("):
        #         variable = variable.split("(")[0]
        #     if variable.count("="):
        #         variable = variable.split("=")[0]
        #
        #     # Small fix to account for a case where at this point we still have parentheses that crash the regex
        #     # TODO: find a better solution
        #     if variable.strip().count("(") or variable.strip().count(")"):
        #         variable = variable.replace("(", "")
        #         variable = variable.replace(")", "")
        #         variable = variable.strip()

            # pattern = "\W*%s\W*" % variable.strip()
            # for _line_index, _line in enumerate(self.lines[_routine_start:_routine_end]):
            #     if not _found:
            #         if _line.count('::'):
            #             continue
            #         if re.search(pattern, _line, re.IGNORECASE):
            #             _used_declarations.append(_all_declaration[index].strip())
            #             _found = True
            #     else:
            #         break
        # if not _used_declarations == _all_declaration:
        #     var_not_used = [var for var in _all_declaration if not _used_declaration.count(var)]
        #     warnings.warn("Routine not found in vault database %s" % var_not_used)
        return _all_declaration

    # Definition: Given the routine name, the name of the dummy arguments and the lines containing declarations,
    # this method finds the attributes of each dummy arguments and adds all the information
    # into the object variable "routines".
    # Input:
    #       -routine_name: string with the routine name
    #       -dummy_arguments: list of strings with argument names
    #       -declarations: list of strings with variable declarations
    def add_struct_members(self, file, structure):
        try:
            for obj in file.objects:
                if obj.name == "main":
                    main = obj
                    break
            for member in structure.variables:
                main.add_variable(member)

        except UnboundLocalError:
            print("Main not found for structure %s", structure.name)

    def find_procedure_variables(self, routine_name, dummy_arguments, declarations, header):
        if header.count("function"):
            routine_or_function = "function"
        elif header.count("subroutine"):
            routine_or_function = "routine"
        else:
            raise ValueError("Header does not correspond to function or routine")

        if routine_or_function == "routine":
            obj_type = Routine
        elif routine_or_function == "function":
            obj_type = Function
        else:
            raise ValueError("routine_or_function must be 'routine' or 'function'")

        ob = obj_type(routine_name, self.module)

        # Reverse approach
        dummy_arguments = [x.lower() for x in dummy_arguments]
        for declaration in declarations:
            pattern = "(.*)::(.*)"
            m = re.search(pattern, declaration)
            # In this case is not a variable declaration
            if declaration.count("contiguous"):
                continue
            _attributes, variable = m.groups()
            # Check if the dummy argument is inside the declaration arguments
            # If it is, take the type, dimension, intent and optional
            _var_type, _dimension, _intent, _optional, \
            _allocatable, pointer, _is_parameter = parse_attributes(_attributes, variable)

            variable_name = split_declaration_arguments(variable)

            if len(variable_name) != 1:
                raise AssertionError("Problem identifying variable declaration.")
            variable_name = variable_name[0]
            _v = Variable(ob, variable_name, _var_type, _dimension, self.module)
            _v.allocatable = _allocatable
            _v.is_parameter = _is_parameter
            _v.pointer = pointer
            if variable_name.lower() in dummy_arguments:
                _v.intent = _intent
                _v.is_optional = _optional
                _v.is_dummy_argument = True
            ob.add_variable(_v)

        # Save the position of the argument
        found_vars = [variable.name.lower() for variable in ob.variables]

        for index, dummy_argument in enumerate(dummy_arguments):
            if dummy_argument in found_vars:
                # TODO: save position of already found vars
                ob.get_variable_by_name(dummy_argument).position = index  # variable.position = index
                continue
                # If it's not found its an implicitly declared variable
            if dummy_argument[0] in "ijklmn":
                var_type = "integer"
            else:
                var_type = working
            dimension = 0
            intent = None
            optional = None
            _v = Variable(ob, dummy_argument, var_type, dimension, self.module)
            _v.intent = intent
            _v.is_optional = optional
            _v.is_dummy_argument = True
            _v.position = index
            _v.is_parameter = False
            ob.add_variable(_v)
        # Save object
        if routine_or_function == "function":

            ret_type, ret_var = function_type_from_header(header)

            if ret_type:
                ob.type = ret_type
                if not ret_type == 'integer':
                    _v = Variable(ob, ret_var, ret_type, 0, self.module)
                    ob.add_variable(_v)
            else:
                for variable in ob.variables:
                    if variable.name.lower() == ret_var.lower():
                        ob.type = variable.type
                        break
        self.objects.append(ob)

    # Definition: Looks into the sourcefile for routine declarations, storing the name, type,
    # dimensions and optionality of the arguments.
    def find_routines(self):

        # Look into each line
        for _index, line in enumerate(self.lines):
            # Look if routine starts with (recoursive) subroutine
            if re.search("^\s*\w*\s*subroutine", line, re.IGNORECASE):
                if line.lower().strip().find("end") == 0:
                    continue
                routine_start = _index
                # Look for end subroutine statements
                routine_end = self.find_next_routine_end(_index)

                # Find name and dummy arguments
                routine_name, dummy_arguments = self.get_dummy_arguments(_index)
                # Find declarations
                declarations = self.find_routine_declarations(routine_start, routine_end)
                header = self.lines[routine_start].lower()
                # Parse everything and store
                self.find_procedure_variables(routine_name.lower(), dummy_arguments, declarations, header)

    # ERROR: Definition: Looks into the sourcefile for routine declarations, storing the name, type,
    # dimensions and optionality of the arguments.
    def find_types(self):
        from AutoRPE.UtilsRPE.Functions import pattern_type, pattern_extended_type, pattern_end_type
        # Look into each line
        in_type_declaration = False
        types = {}
        type_name = None
        variable_counter = None
        for _index, line in enumerate(self.lines):
            # Look if routine starts with subroutine
            if not in_type_declaration:
                # check if a type has been declared
                m = re.search(pattern_type, line, re.IGNORECASE)
                if m:
                    in_type_declaration = True
                    variable_counter = 0
                    type_name = m.groups(0)[0]
                    types[type_name] = {}
                    ob = DerivedType(type_name, self.module)
                    self.objects.append(ob)
                # Case in which a type is extending another type
                match_extend = re.search(pattern_extended_type, line, re.IGNORECASE)
                if match_extend:
                    in_type_declaration = True
                    variable_counter = 0
                    type_name = match_extend.groups()[2]
                    extended_type = match_extend.groups()[1]
                    types[type_name] = {}
                    ob = DerivedType(type_name, self.module, extended_type)
                    self.objects.append(ob)

            else:
                # Check if we are exiting a type declaration
                m = re.search(pattern_end_type, line, re.IGNORECASE)
                if m:
                    in_type_declaration = False
                else:
                    pattern = "(.*)::(.*)"
                    m = re.search(pattern, line)
                    if m:
                        attributes, variables = m.groups()
                        var_type, dimension, intent, optional, \
                        allocatable, pointer, is_parameter = parse_attributes(attributes, variables)

                        variables = split_declaration_arguments(variables)
                        for var in variables:
                            types[type_name][var] = var_type
                            v = Variable(ob, var, var_type, dimension, self.module)
                            v.position = variable_counter
                            v.is_dummy_argument = True
                            v.allocatable = allocatable
                            v.is_parameter = is_parameter
                            v.pointer = pointer
                            v.is_member_of = type_name
                            ob.add_variable(v)
                            variable_counter += 1
        return types

    def find_module_variables(self):
        from AutoRPE.UtilsRPE.Functions import pattern_end_type, pattern_general_type
        main_routine = ModuleMain(self.filename)

        line_to_check = True
        self.objects.append(main_routine)
        pattern = "(.*)::(.*)"
        for line in self.lines:
            if line_to_check:
                # Check if we are entering a type declaration zone
                is_type_declaration = re.search(pattern_general_type, line, re.IGNORECASE)
                if is_type_declaration or is_start_function_statement(line):
                    line_to_check = False
                    continue
            else:
                # Check if we are exiting a type declaration zone
                is_end_type_declaration = re.search(pattern_end_type, line, re.IGNORECASE)
                if is_end_type_declaration or is_end_function_statement(line):
                    line_to_check = True
                # Even if we are inside or exiting a type declaration zone, we have to jump to next line
                continue

            if line.lower().count("contains"):
                break
            if line.lower().count("end program"):
                break
            match = re.search(pattern, line)
            if match:
                attributes, variables = match.groups()
                var_type, dimension, intent, optional, \
                allocatable, pointer, is_parameter = parse_attributes(attributes, variables)
                variables = split_declaration_arguments(variables)
                if var_type:
                    for variable in variables:
                        v = Variable(main_routine, variable, var_type, dimension, self.module)
                        v.allocatable = allocatable
                        v.pointer = pointer
                        v.is_parameter = is_parameter
                        main_routine.add_variable(v)

    # Definition: Looks into the sourcefile for routine declarations, storing the name, type,
    # dimensions and optionality of the arguments.
    def find_functions(self):
        # Look into each line
        for _index, line in enumerate(self.lines):
            if is_start_function_statement(line):
                function_start = _index
                # Look for end subroutine statements
                function_end = self.find_next_function_end(_index)

                # Find name and dummy arguments
                function_name, dummy_arguments = self.get_dummy_arguments(_index)
                # Find declarations
                declarations = self.find_routine_declarations(function_start, function_end)
                # Parse everything and store
                header = self.lines[function_start].lower()
                self.find_procedure_variables(function_name.lower(), dummy_arguments, declarations, header)

    def find_implicit_functions(self, vault):
        try:
            current_block = Current_block(self.module.main())
        except AttributeError:
            return None
        # Definition: Looks into the sourcefile for routine declarations, storing the name, type,
        # dimensions and optionality of the arguments.
        pattern = "^ *(\w+) *\([\w, ]*\) *= *(.*)"
        # Look into each line
        for _index, line in enumerate(self.lines):
            m = re.search(pattern, line)
            if m:
                routine = current_block.routine
                call = close_brackets(line[m.start(1):])
                call_arguments = find_call_arguments(call)
                is_function = False
                for argument in call_arguments:
                    argument_type = type_of_contents(argument, routine=routine, vault=vault)
                    if argument_type != "integer":
                        is_function = True
                        break

                if is_function:
                    function_name = find_called_function(call)
                    new_function = Function(function_name, module=self.module)
                    new_function.type = type_of_contents(function_name, routine=routine, vault=vault)
                    for argument_index, argument in enumerate(call_arguments):
                        var = routine.get_variable_by_name(argument)
                        var.is_dummy_argument = True
                        var.position = argument_index
                        var.intent = "in"
                        new_function.add_variable(var)
                    self.objects.append(new_function)
            else:
                current_block = in_which_block(line, current_block=current_block, vault=vault)

    # Definition: Looks into the sourcefile for interface declarations, storing the name.
    def find_interfaces(self, object_procedures):
        for _index, _line in enumerate(self.lines):
            if _line.strip().lower().find("interface") == 0:
                interface_start = _index
                interface_end = self.find_next_interface_end(_index)
                try:
                    interface_name = _line.split()[1].strip()
                except IndexError:
                    # interface to an external procedure or a procedure used in a procedure with the DEFERRED attribute
                    continue
                routines = []
                line_routines = None
                for interface_line in self.lines[interface_start:interface_end + 1]:
                    if interface_line.strip().lower().find("module procedure") == 0:
                        line_routines = interface_line.lower().replace("module procedure", "")
                        line_routines = [_r.strip() for _r in line_routines.split(",")]
                        routines += line_routines
                    elif interface_line.strip().lower().strip().find("subroutine") == 0:
                        routine_name = interface_line[:interface_line.find("(")].split()[1]
                        routines += [routine_name]

                interface_procedures = [object_procedures[_routine] for _routine in routines]
                ob = Interface(interface_name)
                for procedure in interface_procedures:
                    ob.add_procedure(procedure)
                if not ob.procedures:
                    raise AssertionError("No routine found for interface %s" % interface_name)
                self.objects.append(ob)

    def find_used_modules(self, vault):
        current_block = Current_block(self.module.main())
        for line in self.lines:
            current_block = in_which_block(line, current_block, vault=vault)
            line = line.strip().lower()
            # No need to parse the entire file
            if line.count("implicit none"):
                break
            if line.find("use ") == 0:
                module_name = line.split()[1]
                if module_name.count(","):
                    module_name = module_name.replace(",", "")
                for module in vault.modules:
                    if module.name == module_name:
                        self.module.uses.append(module)

                # Look for
                pattern = "(\w+) *=> *(\w+)"
                matches = re.findall(pattern, line)
                if matches:
                    for match in matches:
                        pointer, target = match
                        target_var = self.module.uses[-1].main().get_variable_by_name(target)
                        new_var = Variable(current_block.routine, pointer, target_var.type, target_var.dimension)
                        current_block.routine.add_variable(new_var)


class ImplementRPE(ReadSourceFile):
    def __init__(self, _filename, vault):
        super(ImplementRPE, self).__init__(_filename)

        # Dictionaries and lists about source info
        self.source_info = vault
        if self.source_info:
            module_name = _filename.replace(".f90", "")
            if module_name.count("/"):
                module_name = module_name.split("/")[-1]
            self.module = self.source_info[module_name]
        else:
            self.module = None

    # Definition:  given a file, puts back on the same line commands split with a &
    def merge_lines(self):
        for _index, _line in enumerate(self.lines):
            line = self.lines[_index]
            if line.count("&"):
                i = 1
                while line and line.strip()[-1] == "&":
                    line1 = line
                    line2 = self.lines[_index + i]
                    line = line1.replace("&", "").rstrip() + line2

                    i += 1
                line = line.replace("&", "")
                line = re.sub(", +", ", ", line)
                line = re.sub(":  +", ":  ", line)
                self.lines[_index] = line
                for j in range(1, i):
                    self.lines[_index + j] = ""
        # now the file is full of empty lines
        self.remove_empty_lines()

    # Split lines which contains more than one command
    # TODO: We should be sure that we are not splitting by characters in literals
    def split_lines(self):

        _index = 0
        length = len(self.lines)
        while _index < length:
            line = self.lines[_index]
            if line.count(";"):
                parts = split_semicolon(line)
                self.lines[_index] = parts[0]
                for i in range(1, len(parts)):
                    self.lines.insert(_index + 1, parts[i])
                    length += 1
                    _index += 1
            _index += 1
        self.rebuild_text()

    # Split declarations of variables putting a single variable in each line
    def split_declarations(self):
        _index = 0
        length = len(self.lines)
        while _index < length:
            line = self.lines[_index]
            line = remove_literals(line)
            if line.count("::"):
                attributes, arguments = line.split("::")
                arguments = find_call_arguments("c(%s)" % arguments)
                if len(arguments) > 1:
                    new_line = "%s :: %s" % (attributes, arguments[0])
                    self.lines[_index] = new_line
                    for argument in arguments[1:]:
                        new_line = "%s :: %s" % (attributes, argument)
                        self.lines.insert(_index + 1, new_line)
                        length += 1
                        _index += 1

            _index += 1
        self.rebuild_text()

    # Remove comments from a file
    def remove_all_comments(self):
        for _index, _line in enumerate(self.lines):
            self.lines[_index] = remove_comments(_line)

    # Remove empty lines from a file
    def remove_empty_lines(self):
        lines = [line for line in self.lines if line.strip()]
        self.lines = lines

    # If parameter declaration is made in Fortran 77 style
    def remove_F77_parameter_declaration(self):

        pattern = '^\s*PARAMETER\s*\(\s*[\w+=\s*]'

        for line_index, line in enumerate(self.lines):

            # Search for lines with the 'PARAMETER' statement
            match = re.match(pattern, line, re.I)
            if match:
                param_line = line.replace(")", "")

                # Store parameter names
                parameters_name = [parameter.split("=")[0].strip() for parameter in
                                   param_line.split("(")[1].split(",")]

                # Store parameter values
                parameters_val = [parameter.split("=")[1].strip() for parameter in
                                  param_line.split("(")[1].split(",")]

                for parameter_name, parameter_value in zip(parameters_name, parameters_val):

                    # Search for each parameter declaration backward in the file
                    for dec_line_index, dec_line in enumerate(reversed((self.lines[:line_index]))):
                        parameter_pattern = '.*::\s*' + parameter_name + '$'
                        param_declaration = re.match(parameter_pattern, dec_line, re.I)
                        if (line_index - dec_line_index - 1) == 0:
                            raise AssertionError("declaration of parameter %s not found in file %s" % (
                                parameter_name, self.filename))
                        # Put back in place the parameter attribute and assign the parameter vale
                        if param_declaration:
                            # Add param statement
                            piece_to_replace = dec_line.split('::')[0]
                            replacement = "%s, parameter " % piece_to_replace
                            dec_line = dec_line.replace(piece_to_replace, replacement)

                            # Add param assignation
                            dec_line = dec_line + "=" + parameter_value

                            # Substitute the line
                            self.lines[line_index - dec_line_index - 1] = dec_line

                            break

                # Delete parameter declaration
                self.lines[line_index] = ""

    def rebuild_text(self):
        self.text = "".join([l + "\n" for l in self.lines[:]])

    def save_modified_file(self, filepath):
        with open(filepath, "w") as output_file:
            output_file.write(self.text)

    def pre_process(self, replace_reals=False, replace_wrk_alloc=False):
        # Remove all comments from all the lines
        self.remove_all_comments()
        # Deletes all the empty lines
        self.remove_empty_lines()
        # Merges commands split on multiples lines
        self.merge_lines()
        # Split multiple command lines
        self.split_lines()
        # Split lines with more than one declaration
        self.split_declarations()

        # Just needed for old Nemo version
        if replace_wrk_alloc:
            self.replace_workalloc_in_file()

        # remove parameter assignation F77 style
        self.remove_F77_parameter_declaration()
        # Put text back together
        self.rebuild_text()
        if replace_reals:
            # Replace real declarations with type(rpe_var)
            self.replace_real_declaration()
            # Workaround for parameters
            self.replace_parameter_declaration()

    def post_process(self):
        # If we don't have source info there's nothing to process
        if self.source_info is None:
            raise AssertionError("It is needed to have a vault database to postprocess the code")

        # Process intrinsic calls
        self.fix_intrinsic_calls()
        # Prosess intrinsics function
        self.fix_implicit_functions()
        # Process routine calls
        self.fix_routines()
        # Process function calls
        self.fix_functions()

        self.fix_type_conversion()

        self.fix_data_statement()

        # Process read and write calls
        self.fix_read_and_write_calls()
        # Fix pointer assignations
        self.fix_pointer_assignations()
        # Add use rp_emulator statement
        self.add_use_rpe_statement()

    def fix_pointer_assignations(self):
        current_block = Current_block(self.module.main())
        pattern = "^[^:'\n\w]*([\w%\(\):,]+) *=> *([\w%\(\):, -]*)"

        for line_index, line in enumerate(self.lines):
            if not re.search(pattern, line):
                current_block = in_which_block(line, current_block, vault=self.source_info)
            else:
                m = re.search(pattern, line)
                pointer, target = m.groups(0)
                pointer_type = type_of_contents(pointer, routine=current_block.routine, vault=self.source_info)
                target_type = type_of_contents(target, routine=current_block.routine, vault=self.source_info)
                if pointer_type in real_types and target_type == "rpe_var":
                    self.lines[line_index] = line.replace(target, "%s%%val" % target)

    # Look at each line for REAL variable declarations and replace it with type(rpe_var)
    def replace_real_declaration(self):
        for _index, line in enumerate(self.lines):
            _match = re.search('[^\!]*(real).*(::|(function))', line, flags=re.I)
            if _match:
                _match_real_with_prec = re.search('[^\!]*(real *\([\w= ]*\)).*(::|(function))', line, flags=re.I)
                if _match_real_with_prec:
                    real_type = _match_real_with_prec.groups(0)[0]
                    line = line.replace(real_type, "type(rpe_var)")
                    self.lines[_index] = line
                else:
                    _match = re.search('^\s*\w?(real).*(::|(function))', line, flags=re.I)
                    if _match:
                        real_type = _match.groups(0)[0]
                    else:
                        raise Exception
                    line = line.replace(real_type, "type(rpe_var)")
                    self.lines[_index] = line
        self.rebuild_text()

    def replace_parameter_declaration(self):
        # If there's an assignation whe have to find a workaround
        for _index, line in enumerate(self.lines):
            if re.search('^[^\!]*type\(rpe_var\).*::.*=', line, flags=re.I):
                arguments = line.split("::")[1]

                type_of_exception = find_kind_of_parameter_exception(arguments)

                # The assignation is made through a real number
                if type_of_exception == 'is_a_number':
                    piece_to_replace = arguments.split("=")[1].strip()
                    replacement = " rpe_var( 52, %s ) " % piece_to_replace
                    self.lines[_index] = line.replace(piece_to_replace, replacement)
                # The assignation is made with a RPE variable
                elif type_of_exception == "regular":
                    piece_to_replace = arguments.split("=")[1].strip()
                    # take the real part of the RPE var
                    piece_to_replace_real = " %s%%val " % piece_to_replace
                    replacement = " rpe_var( 52, %s ) " % piece_to_replace_real
                    self.lines[_index] = replace_variable_exact_match(piece_to_replace, replacement, line)
                elif type_of_exception == "multiple_variables":
                    for argument in arguments.split(","):
                        piece_to_replace = argument.split("=")[1].strip()
                        replacement = " rpe_var(  52, %s ) " % piece_to_replace
                        line = line.replace(piece_to_replace, replacement)
                    self.lines[_index] = line
                elif type_of_exception == "array_initialization":
                    # Keep just the right part of  assignation avoid error when dimensions are specified
                    piece_to_replace = arguments.split("=")[1].strip()
                    replacement = add_rpe_var_to_arguments(piece_to_replace)

                    self.lines[_index] = line.replace(piece_to_replace, replacement)
                elif type_of_exception == "array_reshape":
                    array = arguments.split("[")[1]
                    array = array.split("]")[0]

                    # List the unique arguments inside the reshape function
                    arguments_in_reshape = list(set([a.strip() for a in array.split(",")]))

                    # Add the rpe_var cast into each one of them
                    for arg in arguments_in_reshape:
                        piece_to_replace = arg.strip()
                        replacement = " rpe_var(  52, %s ) " % piece_to_replace
                        line = line.replace(piece_to_replace, replacement)
                    self.lines[_index] = line
                elif type_of_exception == "has_operation":
                    piece_to_replace = arguments.split("=")[1].strip()
                    # Get the elements in operation
                    split_elements = split_elements_in_operations(piece_to_replace)
                    replacement = " rpe_var(  52, %s ) " % piece_to_replace
                    for element in split_elements:
                        operand = element.strip()

                        # If is a real
                        if not is_a_number(operand):
                            operands_real_part = " %s%%val " % operand
                            replacement = replace_variable_exact_match(operand, operands_real_part, replacement)
                    line = line.replace(piece_to_replace, replacement)
                    self.lines[_index] = line
                elif type_of_exception == 'has_intrinsics':

                    # Get the name of the intrinsic to split the line
                    intrinsic_name = has_intrinsics(arguments, True)
                    piece_to_replace = arguments.lower().split(intrinsic_name)[1].strip()

                    # Match the name of the RPE variable inside the parenthesis
                    p = re.search('\((.*?)\)', piece_to_replace)
                    piece_to_replace = p.group(1)

                    # Get the real part of the RPE variable
                    replacement = " %s%%val " % piece_to_replace

                    # Substitute the mach back in the line
                    line = line.replace(piece_to_replace, replacement)
                    piece_to_replace = line.split("=")[1].strip()

                    # Add the cast to RPE var
                    replacement = " rpe_var( 52, %s ) " % piece_to_replace

                    # Save changes
                    self.lines[_index] = line.replace(piece_to_replace, replacement)
                    # self.lines[_index] = line
                elif type_of_exception == "already_fixed":
                    continue
        self.rebuild_text()

    def fix_data_statement(self):
        changed_something = False
        current_block = Current_block(self.module.main())
        # If there's an assignation whe have to find a workaround
        for _index, line in enumerate(self.lines):
            match = re.search(r'^( *)data +(.*) +/(.*)/', line, flags=re.I)
            if match:
                spaces, var_name, values = match.groups(0)
                if var_name.count(","):
                    var_type = type_of_contents(var_name.split(",")[0], routine=current_block.routine,
                                                vault=self.source_info)
                else:
                    var_type = type_of_contents(var_name, routine=current_block.routine, vault=self.source_info)
                if var_type != "rpe_var":
                    continue
                values = ",".join(["rpe_var(52, %s)" % x for x in values.split(",")])
                self.lines[_index] = "%sdata %s /%s/" % (spaces, var_name, values)
                changed_something = True
            else:
                current_block = in_which_block(line, current_block, vault=self.source_info)
        if changed_something:
            self.rebuild_text()

    def add_use_rpe_statement(self):
        # Adding module rp_emulator (in theory adding this to par_kind.f90 should be enough,
        # but theory fails and compiler complains so I'll add it everywhere
        self.lines.insert(1, "USE rp_emulator")
        self.rebuild_text()

    def fix_read_and_write_calls(self):
        self.fix_read_calls()
        self.fix_write_calls()

    def fix_read_calls(self):
        current_block = Current_block(self.module.main())
        pattern = "[^\w](read *\(.*)"
        for line_index, line in enumerate(self.lines):
            matches = re.search(pattern, line, re.IGNORECASE)
            if re.search("write *\(", line, re.I):
                continue
            if matches:
                read_call = close_brackets(matches.groups(1)[0])
                # Remove the read call
                read_arguments = line[matches.start(1):].replace(read_call, "")
                # Continue just if arguments follow the call
                if not read_arguments:
                    continue
                replacement = read_arguments.replace(" ", "")
                arguments = split_read_arguments(replacement)

                # Check the argument type and replace just if is an rpe
                for argument_index, argument in enumerate(arguments):
                    argument_type = type_of_contents(argument, routine=current_block.routine, vault=self.source_info)
                    if argument_type == "rpe_var":
                        replacement = replace_variable_exact_match(argument, argument + "%val", replacement)

                # Rebuilt the line
                # self.lines[line_index] = read_call + " " + read_arguments
                self.lines[line_index] = line.replace(read_arguments, replacement)
            else:
                current_block = in_which_block(line, current_block, vault=self.source_info)

    def fix_write_calls(self):
        current_block = Current_block(self.module.main())
        pattern = "^[^\!]*\s*write\([^()]*\) (.*)."

        for line_index, line in enumerate(self.lines):
            if re.search("WRITE *\(", line):
                wc = close_brackets(line[line.find("WRITE"):])

                line = line.replace(wc, "WRITE()")
            else:
                current_block = in_which_block(line, current_block, vault=self.source_info)
            matches = re.search(pattern, line.strip(), re.IGNORECASE)
            if matches:
                line_container, arguments = separate_line_container_and_arguments(self.lines[line_index])
                if not arguments:
                    continue
                mask = literal_mask(line)
                for argument_index, argument in enumerate(arguments):
                    if mask[line.index(argument) + 1]:
                        continue
                    argument_type = type_of_contents(argument, routine=current_block.routine, vault=self.source_info)
                    if argument_type == "rpe_var":
                        arguments[argument_index] = "real( %s )" % argument
                replacement = ",".join(arguments)
                self.lines[line_index] = line_container.replace("###TOREPLACE###", replacement)

    def fix_routines(self):
        current_block = Current_block(self.module.main())
        exceptions = {}
        temporal_variable_counter = Counter()
        for _index, _line in enumerate(self.lines):
            # Find routine calls in line
            routine_calls = calls_to_routines(_line)
            if not routine_calls:
                # If not, look if the code is entering or exiting a routine block
                current_block = in_which_block(_line, current_block, vault=self.source_info)
            else:
                temporal_line = _line
                for call in routine_calls:
                    is_external = False
                    called_routine = find_called_routine(call)
                    called_arguments = find_call_arguments(call)
                    if not called_arguments:
                        continue

                    # In case the procedure is not found, consider it external
                    try:
                        procedure = self.source_info.get_procedure(called_routine)
                        if isinstance(procedure, Routine):
                            pass
                        elif isinstance(procedure, Interface):
                            procedure = match_arguments_and_interface(called_arguments, interface=procedure,
                                                                      routine=current_block.routine,
                                                                      vault=self.source_info)
                    except Error.ProcedureNotFound:
                        warnings.warn("Routine not found in vault database %s" % called_routine)
                        is_external = True

                        # continue

                    temporal_call = call
                    if not is_external:
                        for argument_index, argument in enumerate(called_arguments):
                            if is_keyword_argument(argument):
                                [keyword, val] = argument.lower().split("=", 1)
                                argument_type = type_of_contents(val.strip(), routine=current_block.routine,
                                                                 vault=self.source_info)
                                dummy_argument = procedure.get_variable_by_name(keyword.strip())
                                if not dummy_argument:
                                    raise AssertionError("Variable not found: %s" % keyword)
                                if argument_type == dummy_argument.type:
                                    continue
                                elif argument_type in real_types and dummy_argument.type == "rpe_var":
                                    if dummy_argument.intent == "in":
                                        temporal_call = add_rpe_to_argument(temporal_call, argument_index)
                                    else:
                                        raise AssertionError("Unknwon intent: %s" % dummy_argument.intent)
                                elif argument_type == "rpe_var" and dummy_argument.type in real_types:
                                    if dummy_argument.intent == "in":
                                        temporal_call = add_real_to_argument(temporal_call, argument_index)
                                    elif dummy_argument.intent == "out":
                                        temporal_call = add_val_to_argument(temporal_call, argument_index)
                                    else:
                                        raise AssertionError("Unknwon intent: %s" % dummy_argument.intent)
                            else:
                                argument_type = type_of_contents(argument, routine=current_block.routine,
                                                                 vault=self.source_info)
                                dummy_argument = procedure.get_variable_by_position(argument_index)
                                if not dummy_argument:
                                    raise AssertionError("Variable not found in position: %i" % argument_index)
                                if argument_type == dummy_argument.type:
                                    continue
                                elif argument_type in real_types and dummy_argument.type == "rpe_var":
                                    if dummy_argument.intent == "in":
                                        temporal_call = add_rpe_to_argument(temporal_call, argument_index)
                                    else:
                                        try:
                                            exceptions[_index].append(
                                                (argument, argument_index, dummy_argument, current_block.routine))
                                        except KeyError:
                                            exceptions[_index] = [
                                                (argument, argument_index, dummy_argument, current_block.routine)]
                                elif argument_type == "rpe_var" and dummy_argument.type in real_types:
                                    if dummy_argument.intent == "in":
                                        temporal_call = add_real_to_argument(temporal_call, argument_index)
                                    elif dummy_argument.intent == "out":
                                        temporal_call = add_val_to_argument(temporal_call, argument_index)
                                    elif dummy_argument.intent == "inout":
                                        temporal_call = add_val_to_argument(temporal_call, argument_index)
                                    else:
                                        raise AssertionError("Unknwon intent: %s" % dummy_argument.intent)
                        temporal_line = temporal_line.replace(call, temporal_call)
                    else:
                        for argument_index, argument in enumerate(called_arguments):
                            # TODO: Probably should find a better solution, the code should not identify an array as a
                            # Todo: ... function even if the name coincides.
                            # if argument == ":":
                            #     continue
                            argument_type = type_of_contents(argument, routine=current_block.routine,
                                                             vault=self.source_info)
                            if argument_type == "rpe_var":
                                # TODO: Probably I should check first if the argument is a single variable and can be used as an output argument or not
                                if re.match("^\w+$", argument):
                                    temporal_call = add_val_to_argument(temporal_call, argument_index)
                                elif re.match("^\w+ *\([^\(\)]*\)$", argument):
                                    temporal_call = add_val_to_argument(temporal_call, argument_index)
                                elif not has_operations(argument):
                                    temporal_call = add_val_to_argument(temporal_call, argument_index)
                                else:
                                    temporal_call = add_real_to_argument(temporal_call, argument_index)
                        temporal_line = temporal_line.replace(call, temporal_call)
                self.lines[_index] = temporal_line
        self.rebuild_text()

        # Manage exceptions
        added_lines = 0
        for line in sorted(exceptions):
            for argument, argument_index, dummy_argument, routine in exceptions[line]:
                current_line = line + added_lines
                use_temporal_variable(self, current_line, argument, argument_index, routine=routine,
                                      dummy=dummy_argument, counter=temporal_variable_counter)
                added_lines += 3
            added_lines += 2 * len(exceptions[line])
        self.rebuild_text()

    def fix_functions(self):
        current_block = Current_block(self.module.main())
        for line_index, line in enumerate(self.lines):
            function_calls = calls_to_functions(line, self.source_info, current_routine=current_block.routine)
            if not function_calls:
                # If not, look if the code is entering or exiting a routine block
                current_block = in_which_block(line, current_block, vault=self.source_info)
            else:
                # Fix function calls (line_index
                temporal_line = line
                for call in function_calls:
                    # Small workaround to avoid identifying routines as functions.
                    patt = "call +%s" % call
                    patt = patt.replace("(", r"\(")
                    patt = patt.replace(")", r"\)")
                    patt = patt.replace("/", r"\/")
                    patt = patt.replace("*", r"\*")
                    m = re.search(patt, line, re.I)
                    if m:
                        continue
                    is_external = False
                    called_function = find_called_function(call)

                    if called_function.lower() == current_block.routine.name.lower():
                        continue
                    called_arguments = find_call_arguments(call)
                    if not called_arguments:
                        continue

                    try:
                        procedure = self.source_info.get_procedure(called_function)
                        if isinstance(procedure, Routine):
                            pass
                        elif isinstance(procedure, Interface):
                            procedure = match_arguments_and_interface(called_arguments, interface=procedure,
                                                                      routine=current_block.routine,
                                                                      vault=self.source_info)
                    except Error.ProcedureNotFound:
                        warnings.warn("Routine not found in vault database %s" % called_function)
                        is_external = True

                    if not is_external:
                        temporal_call = call
                        for argument_index, argument in enumerate(called_arguments):
                            # TODO: Probably should find a better solution, the code should not identify an array as a
                            # Todo: ... function even if the name coincides. sbcice_lim and lib_fortran fail without it
                            if argument == ":":
                                continue
                            if is_keyword_argument(argument):
                                keyword = argument.split("=")[0].strip()
                                argument_type = type_of_contents(argument, routine=current_block.routine,
                                                                 vault=self.source_info)
                                dummy_argument = procedure.get_variable_by_name(keyword)
                                if not dummy_argument:
                                    raise AssertionError("Variable not found: %s" % keyword)
                                if argument_type == dummy_argument.type:
                                    continue
                                elif argument_type in real_types and dummy_argument.type == "rpe_var":
                                    if dummy_argument.intent == "in":
                                        temporal_call = add_rpe_to_argument(temporal_call, argument_index)
                                    else:
                                        raise AssertionError("Unknwon intent: %s" % dummy_argument.intent)
                                elif argument_type == "rpe_var" and dummy_argument.type in real_types:
                                    if dummy_argument.intent == "in":
                                        temporal_call = add_real_to_argument(temporal_call, argument_index)
                                    elif dummy_argument.intent == "out":
                                        temporal_call = add_val_to_argument(temporal_call, argument_index)
                                    else:
                                        raise AssertionError("Unknwon intent: %s" % dummy_argument.intent)
                            else:
                                argument_type = type_of_contents(argument, routine=current_block.routine,
                                                                 vault=self.source_info)
                                dummy_argument = procedure.get_variable_by_position(argument_index)
                                if not dummy_argument:
                                    raise AssertionError("Variable not found in position: %i" % argument_index)
                                if argument_type == dummy_argument.type:
                                    continue
                                elif argument_type in real_types and dummy_argument.type == "rpe_var":
                                    if dummy_argument.intent == "in":
                                        temporal_call = add_rpe_to_argument(temporal_call, argument_index)
                                    else:
                                        raise AssertionError("Unknwon intent: %s" % dummy_argument.intent)
                                elif argument_type == "rpe_var" and dummy_argument.type in real_types:
                                    if dummy_argument.intent == "in":
                                        temporal_call = add_real_to_argument(temporal_call, argument_index)
                                    elif dummy_argument.intent == "out":
                                        temporal_call = add_val_to_argument(temporal_call, argument_index)
                                    elif dummy_argument.intent == "inout":
                                        temporal_call = add_val_to_argument(temporal_call, argument_index)
                                    else:
                                        raise AssertionError("Unknwon intent: %s" % dummy_argument.intent)
                        temporal_line = temporal_line.replace(call, temporal_call)
                    else:
                        temporal_call = call
                        for argument_index, argument in enumerate(called_arguments):
                            # TODO: Probably should find a better solution, the code should not identify an array as a
                            # Todo: ... function even if the name coincides.
                            # if argument == ":":
                            #     continue
                            argument_type = type_of_contents(argument, routine=current_block.routine,
                                                             vault=self.source_info)
                            if argument_type == "rpe_var":
                                # TODO: Probably I should check first if the argument is a single variable and can be used as an output argument or not
                                temporal_call = add_val_to_argument(temporal_call, argument_index)
                        temporal_line = temporal_line.replace(call, temporal_call)
                self.lines[line_index] = temporal_line
        self.rebuild_text()

    def fix_type_conversion(self):
        current_block = Current_block(self.module.main())
        for line_index, line in enumerate(self.lines):
            function_calls = calls_to_procedure(line, procedure_type=DerivedType, vault=self.source_info)
            if not function_calls:
                # If not, look if the code is entering or exiting a routine block
                current_block = in_which_block(line, current_block, vault=self.source_info)
            else:
                # Fix function calls (line_index
                temporal_line = line
                for call in function_calls:
                    called_function = find_called_function(call)
                    called_arguments = find_call_arguments(call)
                    if not called_arguments:
                        continue

                    procedure = self.source_info.get_procedure(called_function)
                    if isinstance(procedure, DerivedType):
                        pass
                    else:
                        raise TypeError("The %s procedure does not have the DerivedType type" % procedure.name)

                    temporal_call = call
                    for argument_index, argument in enumerate(called_arguments):
                        # TODO: Probably should find a better solution, the code should not identify an array as a
                        # Todo: ... function even if the name coincides. sbcice_lim and lib_fortran fail without it
                        if argument == ":":
                            continue
                        argument_type = type_of_contents(argument, routine=current_block.routine,
                                                         vault=self.source_info)
                        dummy_argument = procedure.get_variable_by_position(argument_index)
                        if not dummy_argument:
                            raise AssertionError("Variable not found in position: %i" % argument_index)
                        if argument_type == dummy_argument.type:
                            continue
                        elif argument_type in ["single", "double"] and dummy_argument.type == "rpe_var":
                            temporal_call = add_rpe_to_argument(temporal_call, argument_index)
                        else:
                            raise AssertionError("Types do not coincide and the case is not considered %s %s" % (
                                argument_type, dummy_argument.type))
                    temporal_line = temporal_line.replace(call, temporal_call)
                self.lines[line_index] = temporal_line
        self.rebuild_text()

    def fix_intrinsic_calls(self):
        current_block = Current_block(self.module.main())
        for _index, _line in enumerate(self.lines):

            # Cycle through the matches, that can be more than one in a line
            replacements = []
            pieces_to_replace = []
            matches_start_list = []
            mask = literal_mask(_line)
            for intrinsic in list_of_intrinsics:

                # Search for the instrinsic
                fun = r"[ ,(+-=*](%s *)\(" % intrinsic

                matches = re.finditer(fun, _line, flags=re.I)
                matches_start = [m.start(1) for m in matches]

                if matches_start:
                    for start in matches_start:
                        if not mask[start]:
                            matches_start_list.append(start)

            if not matches_start_list:
                current_block = in_which_block(_line, current_block, vault=self.source_info)
                continue
            else:
                matches_start_list.sort(reverse=True)
                for start in matches_start_list:
                    # Save the piece that need change
                    pieces_to_replace.append(_line[start:].strip())

                    # Remove brackets if they surrounds the argument
                    i_match = len(pieces_to_replace) - 1
                    if pieces_to_replace[i_match][0] == "(":
                        pieces_to_replace[i_match] = pieces_to_replace[i_match][1:]
                    pieces_to_replace[i_match] = close_brackets(pieces_to_replace[i_match])
                    _line = _line[:start] + _line[start:].replace(pieces_to_replace[i_match],
                                                                  "ALREADY_FIXED_" + str(i_match))
                    # Split the calls argument
                    number_of_arguments, intervals = find_argument_intervals(pieces_to_replace[i_match])
                    # Vector that will contain the changed pieces
                    replacements.append(pieces_to_replace[i_match])

                    for arg_id in range(number_of_arguments):
                        argument = pieces_to_replace[i_match][intervals[arg_id][0]:intervals[arg_id][1]]
                        if is_a_number(argument):
                            continue
                        # Split the operation in the argument
                        if has_operations(argument):
                            split_elements = split_elements_in_operations(argument)
                            for element in set(split_elements):
                                if element.count("ALREADY_FIXED_"):
                                    continue
                                # TODO: Fix this workaround
                                # The original approach is failing because it is looking for a structure member of
                                # a custom derived type that is declared outside our sources.
                                # Adding error handling to bypass this.
                                try:
                                    argument_type = type_of_contents(element, routine=current_block.routine,
                                                                     vault=self.source_info)
                                except Error.MemberNotFound:
                                    argument_type = None
                                if argument_type == 'rpe_var':
                                    if is_call_to_function(element):
                                        replacements[i_match] = replace_variable_exact_match(element,
                                                                                             "real(" + element + ")",
                                                                                             replacements[i_match])
                                    else:
                                        replacements[i_match] = replace_variable_exact_match(element,
                                                                                             element + "%val",
                                                                                             replacements[i_match])
                        elif argument.count("="):
                            [key, val] = argument.lower().split("=", 1)

                            if key.strip() == 'mask':
                                # Why is not split_element_in_comparison working?
                                # raise AssertionError("Check this is working")
                                values = split_elements_in_comparison(val)
                                if values is not None:
                                    continue
                        else:
                            if argument.count("ALREADY_FIXED_"):
                                continue
                            argument_type = type_of_contents(argument, routine=current_block.routine,
                                                             vault=self.source_info)
                            if argument_type == 'rpe_var':
                                replacements[i_match] = replace_variable_exact_match(argument, argument + "%val",
                                                                                     replacements[i_match])

            # for i_match, replacement in enumerate(replacements):
            for i_match, (piece_to_replace, replacement) in reversed(
                    list(enumerate(zip(pieces_to_replace, replacements)))):
                _line = _line.replace("ALREADY_FIXED_" + str(i_match), piece_to_replace)
                if not pieces_to_replace == replacements:
                    # Starting from the end of the line, and substituting just first match we avoid re-substituting matches
                    _line = _line.replace(piece_to_replace, replacement, 1)
            self.lines[_index] = _line

    def fix_implicit_functions(self):
        try:
            current_block = Current_block(self.module.main())
        except AttributeError:
            return None
        # Definition: Looks into the sourcefile for routine declarations, storing the name, type,
        # dimensions and optionality of the arguments.
        pattern = r"^ *(\w+) *\([\w, ]*\) *= *(.*)"
        # Look into each line
        for _index, line in enumerate(self.lines):
            m = re.search(pattern, line)
            if m:
                routine = current_block.routine
                call = close_brackets(line[m.start(1):])
                call_arguments = find_call_arguments(call)
                is_function = False
                for argument in call_arguments:
                    argument_type = type_of_contents(argument, routine=routine, vault=self.source_info)
                    if argument_type != "integer":
                        is_function = True
                        break

                if is_function:
                    function_name = find_called_function(call)
                    new_function = Function(function_name, module=self.module)
                    new_function.type = type_of_contents(function_name, routine=routine, vault=self.source_info)
                    for argument_index, argument in enumerate(call_arguments):
                        var = routine.get_variable_by_name(argument)
                        var.is_dummy_argument = True
                        var.position = argument_index
                        var.intent = "in"
                        new_function.add_variable(var)

                    self.insert_function(new_function, _index)
                    # Now remove var from vault mutable variables
                    for var in self.source_info.variables:
                        name = var.name
                        module_name = var.routine.module.name
                        routine_name = var.routine.name
                        if name == function_name and module_name == new_function.module.name and routine_name == routine.name:
                            self.source_info.variables.remove(var)
                            break

            else:
                current_block = in_which_block(line, current_block=current_block, vault=self.source_info)

    # Transform an implicit function in a normal function
    def insert_function(self, new_function, line_index):

        function_declaration, function_definition = self.lines[line_index].split("=")

        # Now that we stored the function we can delete its usage
        self.lines[line_index] = ""

        # Now we proceed to remove the variables declaration and the function declaration
        n_var = len(new_function.variables) + 1
        for _index, _line in enumerate(reversed((self.lines[:line_index]))):
            if n_var == 0:
                break
            if _line.count("::"):
                declared_var = _line.split("::")[1].strip()
                for var in new_function.variables:
                    if var.name == declared_var or new_function.name == declared_var:
                        self.lines[line_index - _index - 1] = ""
                        n_var -= 1
                        break
        # Now remove these lines
        ImplementRPE.remove_empty_lines(self)

        # Search for a line to insert the new function definition
        fun_end = ObtainSourcefileInfo.find_next_end(self, line_index, "end subroutine")
        sbr_ned = ObtainSourcefileInfo.find_next_end(self, line_index, "end function")
        insertion_line = min(fun_end, sbr_ned) + 1
        # Insert the function

        self.lines.insert(insertion_line, "END FUNCTION")
        function_definition = "x =" + function_definition
        self.lines.insert(insertion_line, function_definition)
        if new_function.type == "rpe_var":
            ret_val_declaration = "type(" + new_function.type + ") :: x"
        else:
            ret_val_declaration = new_function.type + " :: x"
        self.lines.insert(insertion_line, ret_val_declaration)
        for var in new_function.variables:
            if var.type == "rpe_var":
                var_declaration = "type(" + var.type + ") :: " + var.name
            else:
                var_declaration = var.type + " :: " + var.name
            self.lines.insert(insertion_line, var_declaration)

        function_declaration = "FUNCTION" + function_declaration + "result(x)"
        self.lines.insert(insertion_line, function_declaration)

        self.rebuild_text()


# Miscellaneous Utils
class Counter:
    def __init__(self):
        self.count = 0

    def up(self, how_much=1):
        self.count += how_much
